/*
	MEMORY GAME: Proiect PM 2019, Bucur Adrian-Catalin 336CA
*/

#include <avr/io.h>
#include <util/delay.h>
#include <stdlib.h>
#include <stdio.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include "lcd.h"

#define SHORT_DELAY 500
#define LONG_DELAY 1000
#define TIMER_DELAY 45
#define COUNTDOWN 16
#define HIGHSCORE_SIZE 5
#define read_eeprom_word(address) eeprom_read_word ((const uint16_t*)address)
#define write_eeprom_word(address,value) eeprom_write_word ((uint16_t*)address,(uint16_t)value)
#define update_eeprom_word(address,value) eeprom_update_word ((uint16_t*)address,(uint16_t)value)

int EEMEM  eeprom[HIGHSCORE_SIZE];
int ram[HIGHSCORE_SIZE];
int timer_delay = TIMER_DELAY;
int countdown = COUNTDOWN;
volatile bool timeout;

//INITIALIZARE TIMER0: FOLOSIT PENTRU COUNTDOWN
void timer0_init(void){
    TCCR0A = (1 << WGM01) ;
    TCCR0B =  (1<<CS00) | ( 1<<CS02);
    TCNT0 = 0;
    OCR0A = 255;
    TIMSK0 |= (1 << OCIE0A);
}

//INITIALIZARE TIMER2: FOLOSIT PENTRU SRAND
void timer2_init(){
    TCCR2A=0x00;
	TCCR2B=0x05;
	TCNT2=0x00;
	OCR2A=0x00;
	OCR2B=0x00;
}


//PORNESTE TIMERUL PENTRU COUNTDOWN
void start_countdown(){
	countdown = COUNTDOWN;
	timeout = false;
	timer0_init();
}

//OPRESTE TIMERUL PENTRU COUNTDOWN
void stop_countdown(){
	TCCR0A = 0;
    TCCR0B =  0;
    TCNT0 = 0;
    OCR0A = 255;
    TIMSK0 = 0;
}

//GENEREAZA SUNETE PE BUZZER
void speaker(int beeps){
	for(int i = 0; i < beeps; i++){
		 DDRB |= (1 << PB5);
   		_delay_ms(50);
  		DDRB &= ~(1 <<PB5);
  		_delay_ms(50);
	}
}

//TRATAREA INTRERUPERII PENTRU TIMERUL FOLOSIT LA COUNTDOWN
ISR(TIMER0_COMPA_vect){
    timer_delay--;

    if(timer_delay == 0){

    	//SCHIMB STAREA LED-ULUI SI AFISEZ UN NOU '#'
    	PORTD ^= (1 << PD7);
    	LCD_print("#");

    	timer_delay = TIMER_DELAY;
    	countdown--;

    	//ULTIMELE 5 SECUNDE: BEEP-URI DE AVERTIZARE
    	if(countdown < 5){
    		speaker(1);
    	}

    	//TIMEOUT
    	if(countdown == 0){
    		timeout = true;
    		LCD_writeInstr(LCD_INSTR_clearDisplay);
			LCD_print("Timeout!");
			_delay_ms(SHORT_DELAY);
    		stop_countdown();
    	}
    }
}

//SUNETE DE BUZZER PENTRU CAND USERUL PIERDE
void loose_song(){
	for(int i = 0; i < 4; i++){
		speaker(4);
		 DDRB |= (1 << PB5);
		_delay_ms(100);
		DDRB &= ~(1 <<PB5);
		_delay_ms(100);
		DDRB &= ~(1 <<PB5);
		_delay_ms(200);
		DDRB &= ~(1 <<PB5);
		_delay_ms(50);
	}
}

//PREIA O APASARE DE BUTON DE LA USER
int get_input(){
	while(true){
		if(timeout){
			timeout = false;
			return -1;
		} else if ((PIND & (1 << PD0)) == 0) {
			speaker(1);
			return 1;
		} else if ((PIND & (1 << PD1)) == 0) {
			speaker(2);
			return 2;
		} else if ((PIND & (1 << PD4)) == 0) {
			speaker(3);
			return 3;
		} else if ((PIND & (1 << PD5)) == 0) {
			speaker(4);
			return 4;
		}
	}
}

//PREIA SECVENTA DE LA USER SI O VERIFICA
bool check_passed_level(int *arr, int level){

	start_countdown();
	int user_input;

	for(int i = 0; i < level; i++){

		user_input = get_input();

		if(user_input == -1){
			return false;
		} else if(user_input != arr[i]){
			stop_countdown();
			return false;
		}
	}

	stop_countdown();
	return true;
}

//AFISEAZA SECVENTA PENTRU NIVELUL CURENT
void display_sequence(int *arr, int arr_size){
	char msg[20];

	for(int i = 0; i < arr_size; i++){

		if(arr[i] == 1){
			sprintf(msg, "*****(blue)****");
		} else  if(arr[i] == 2){
			sprintf(msg, "*****(red)*****");
		} else if(arr[i] == 3){
			sprintf(msg, "****(black)****");
		} else {
			sprintf(msg, "****(white)****");
		}

		_delay_ms(SHORT_DELAY);
		LCD_writeInstr(LCD_INSTR_clearDisplay);
		_delay_ms(SHORT_DELAY);
		LCD_print(msg);
		speaker(arr[i]);
	}
}

//VERIFICA DACA SCORUL E IN TOP SI ACTUALIZEAZA IN RAM SI EEPROM
void update_highscores(int level_reached){

	for(int i = 0; i < HIGHSCORE_SIZE; i++){
		ram[i] = read_eeprom_word(&eeprom[i]);
	}

	for(int i = 0; i < HIGHSCORE_SIZE; i++){
		if(level_reached > ram[i]){
			write_eeprom_word(&eeprom[i], level_reached);

			for(int j = i + 1; j < HIGHSCORE_SIZE; j++){
				write_eeprom_word(&eeprom[j], ram[j - 1]);
			}

			LCD_writeInstr(LCD_INSTR_clearDisplay);
			LCD_print("You made the");
			LCD_writeInstr(LCD_INSTR_nextLine);
			LCD_print("highscore list!");

			_delay_ms(LONG_DELAY * 2);

			break;
		}
	}
}


//AFISEAZA LISTA DE HIGHSCORES, ASTEAPTA USER INPUT PT DERULARE
void show_highscores(){

	//CITESC IN RAM VALORILE DIN EEPROM
	for(int i = 0; i < HIGHSCORE_SIZE; i++){
		ram[i] = read_eeprom_word(&eeprom[i]);
	}

	int position = 0;

list:

	//RANDUL DE SUS
	LCD_writeInstr(LCD_INSTR_clearDisplay);
	char msg[16];

    if(position + 1 == 1){
    	sprintf(msg, "1st: Level %d", ram[position]);
    } else if (position + 1 == 2){
    	sprintf(msg, "2nd: Level %d", ram[position]);
    } else if (position + 1 == 3) {
    	sprintf(msg, "3rd: Level %d", ram[position]);
    } else {
    	sprintf(msg, "%dth: Level %d", position + 1, ram[position]);
    }

	LCD_print(msg);
	LCD_writeInstr(LCD_INSTR_nextLine);

	position++;
	position %= HIGHSCORE_SIZE;

	//RANDUL DE JOS
  	if(position + 1 == 1){
    	sprintf(msg, "1st: Level %d", ram[position]);
    } else if (position + 1 == 2){
    	sprintf(msg, "2nd: Level %d", ram[position]);
    } else if (position + 1 == 3) {
    	sprintf(msg, "3rd: Level %d", ram[position]);
    } else {
    	sprintf(msg, "%dth: Level %d", position + 1, ram[position]);
    }
	LCD_print(msg);

	//BUTONUL 1: SCROLL, IN REST EXIT TO MAIN MENU
	int choice = get_input();
	if (choice == 1){
		_delay_ms(50);
		goto list;
	}

}

//INCEPE UN NOU JOC
void play_game(){

	int MAX_LEVEL = 25;
	int arr[MAX_LEVEL];
	int level = 1;

	LCD_writeInstr(LCD_INSTR_clearDisplay);
	LCD_print("NEW GAME");
	_delay_ms(1000);

	while(true){
		char msg[20];

		//AFISEZ NIVELUL CURENT
		_delay_ms(SHORT_DELAY);
		sprintf(msg, "Level %d", level);
		LCD_writeInstr(LCD_INSTR_clearDisplay);
		LCD_print(msg);
		_delay_ms(SHORT_DELAY);

		//GENEREZ SECVENTA RANDOM
		srand(TCNT2);
		for(int i = 0; i < level; i++){
			arr[i] = rand() % 4 + 1;
		}

		display_sequence(arr, level);

		_delay_ms(LONG_DELAY);
		sprintf(msg, "GO!");
		LCD_writeInstr(LCD_INSTR_clearDisplay);
		LCD_print(msg);
		LCD_writeInstr(LCD_INSTR_nextLine);

		//USERUL A PIERDUT: MESAJ SI "MUZICA", UPDATE SCORES
		if(!check_passed_level(arr, level)){
			_delay_ms(LONG_DELAY);
			LCD_writeInstr(LCD_INSTR_clearDisplay);

			sprintf(msg, "Wrong. You lost.");
			LCD_print(msg);
			LCD_writeInstr(LCD_INSTR_nextLine);
			sprintf(msg, "Your score: %d", level - 1);
			LCD_print(msg);

			loose_song();

			update_highscores(level - 1);

			break;
		}

		sprintf(msg, "Win! Next level...");
		LCD_writeInstr(LCD_INSTR_nextLine);
		LCD_print(msg);
		_delay_ms(LONG_DELAY * 1.5);

		level++;
	}
}

int main(void) {

	//INITIALIZARE BUTOANE
	DDRD &= ~(1 << PD0);
	PORTD |= (1 << PD0);

	DDRD &= ~(1 << PD1);
	PORTD |= (1 << PD1);

	DDRD &= ~(1 << PD4);
	PORTD |= (1 << PD4);

	DDRD &= ~(1 << PD5);
	PORTD |= (1 << PD5);

	//INITIALIZARE LED
	DDRD |= (1 << PD7);
	PORTD &= ~(1 << PD7);

	//INITIALIZARE LCD
	LCD_init();
	DDRC |= (1<<PC2);
	PORTC |= (1<<PC2);

	//TIMER1: pentru generare numere random
	sei();
	timer2_init();

	while(1) {

		//MAIN MENU: 2 OPTIUNI
		LCD_writeInstr(LCD_INSTR_clearDisplay);
		LCD_print("Start:    (blue)");
		LCD_writeInstr(LCD_INSTR_nextLine);
		LCD_print("Highscores:(red)");

		int choice = get_input();

		if(choice == 1){
			play_game();
		}

		if(choice == 2){
			show_highscores();
		}
	}

	return 0;
}